/*************************************************************************\
* Copyright (c) 2002 The University of Chicago, as Operator of Argonne
* National Laboratory.
* Copyright (c) 2002 The Regents of the University of California, as
* Operator of Los Alamos National Laboratory.
* This file is distributed subject to a Software License Agreement found
* in the file LICENSE that is included with this distribution. 
\*************************************************************************/
/**** SCCS ID: @(#)libdev.c	1.31 3/25/94 ****/
/*************************************************************************
 * FILE:      libdev.c
 * Author:    Claude Saunders
 * 
 * Purpose:   Library providing I/O and utility functions for accessing
 *            atomic and composite devices. These functions work together
 *            with the device server using RPC.
 * Mods:      login mm/dd/yy: description
 *************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <rpc/rpc.h>
#include <string.h>
#include <math.h>
#include <netdb.h>
#include <unistd.h>
#include "gpdLib2.h"
#include "scan_defs.h"
#include "libdev.h"

/*double atof();*/
double quiet_nan(long n);

/* PRIVATE functions */
static int addDevice(char *dname,char *msg);
static int msgOp(PDEVNODE pdevnode,void *arg);
static int connect_to_server();
static void copyReplynode(PDEVNODE pdevnode,devicelist reply);
static int init_libdev();
static int stringcompare(const void *a, const void *b);
static void *configure_parg(int datatype, int dir, int nargs,
			    char *def, int type, void *arg);

/* PRIVATE data */
static int first_call = 1;           /* for first use initialization */
static void *table;                  /* hash table for device names  */
static CLIENT *cl;                   /* client RPC connection ident  */
static char server[MAXHOSTNAMELEN];  /* server hostname              */
static float timeout = 5.0;          /* global timeout for ca_pend_io*/

/*************************************************************************
 * FUNCTION : devMatch()
 * PURPOSE  : Matches a wildcard pattern against the set of known atomic
 *            and composite device names. Returned is an array of pointers
 *            to characters, one element for each match. The pattern may
 *            contain regular characters and the wildcard characters * and
 *            ?. The matching works the same as UNIX filename completion.
 *            Space is automatically allocated for the array, and sub-
 *            sequently freed up on the next call. The last pointer in
 *            the array is NULL to indicate no more matches.
 * ARGS in  : pattern: wildcard string
 * ARGS out : matches: array of char ptrs, one for each match
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL - success, cnt is valid
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 ************************************************************************/
int devMatch(char *pattern, devnamearr *matches)
{
  namelist *nres;
  namelist list;
  char *localpatt = pattern;
  int nmatches, i;
  static devnamearr namearr = NULL;

  if (first_call) {
    if (init_libdev() != DEV_NORMAL)
      return(DEV_NOSERVER);
    first_call = 0;
  }

  /* Device server performs matching and returns linked list of names */
  if ((nres = device_match_1(&localpatt,cl)) == NULL) {
    clnt_perror(cl,server);
    return(DEV_CALLFAIL);
  }
  /* Free up malloc'ed space from last call to devMatch */
  if (namearr != NULL) {
    int i = 0;
    while (namearr[i] != NULL) {
      free(namearr[i]);
      i++;
    }
    free(namearr);
  }

  /* Count up number of matches by traversing list */
  list = *nres;
  for (nmatches=0 ; list != NULL ; list=list->next, nmatches++);

  /* Allocate space for array and copy server reply */
  namearr = (devnamearr)malloc((nmatches+1)*sizeof(char *));
  *matches = namearr;
  list = *nres;
  for (i=0 ; i < nmatches ; i++) {
    namearr[i] = (char *)malloc(255);
    strcpy(namearr[i],list->dname);
    list = list->next;
  }
  namearr[nmatches] = NULL;    /* last element must be NULL */

  /* perform lexicographic quicksort on pattern matches */
  qsort(namearr, nmatches, sizeof(char *), stringcompare);

  return(DEV_NORMAL);
}

/*************************************************************************
 * FUNCTION : devCount()
 * PURPOSE  : Returns, in cnt, a count of the number of devices in dname. 
 *            An atomic device always returns 1.
 * ARGS in  : dname: name of device (atomic or composite)
 * ARGS out : cnt:   count of number of devices composing dname
 *            names: array of ptrs to atomic device names, 1 for each in dname.
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL - success, cnt is valid
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 *            DEV_NODEV    - dname does not exist
 ************************************************************************/
int devCount(char *dname, int *cnt, devnamearr *names)
{
  char *localname = dname; /* RPC call argument */
  namelist *reply;         /* RPC call return */
  namelist list;
  int i;
  GPDENTRY *pentry;        /* hash table entry */
  PDEVNODE pdevnode;       /* ptr to userPvt of hash table entry */
  static devnamearr namearr = NULL;
  int count = 0;

  if (first_call) {
    if (init_libdev() != DEV_NORMAL)
      return(DEV_NOSERVER);
    first_call = 0;
  }

  *cnt = 0;

  /* Free up malloc'ed space from last call to devCount */
/*
  if (namearr != NULL) {
    int i = 0;
    while (namearr[i] != NULL) {
      free(namearr[i]);
      i++;
    }
    free(namearr);
  }
*/
  
  /* Try to find dname in local hash table */
  pentry = gpdFind(table,dname,"");

  if (pentry == NULL) {         /* can't find, so do RPC call to get info */
    if ((reply = device_count_1(&localname,cl)) == NULL) {
      clnt_perror(cl,server);
      return(DEV_CALLFAIL);
    }
    list = *reply;
    if (list == NULL)
	return(DEV_NODEV);

    for (count=0 ; list != NULL ; list=list->next, count++);

    /* Allocate space for array and copy names */
    namearr = (devnamearr)malloc((count+1)*sizeof(char *));
    *names = namearr;    

    list = *reply;
    for (i=0 ; i < count ; i++) {
      namearr[i] = (char *)malloc(255);
      strcpy(namearr[i], list->dname);
      list = list->next;
    }
  }
  else {                        /* dname in hash table, get count from there */
    pdevnode = (PDEVNODE)pentry->userPvt;
    if (pdevnode->type == COMPOSITE) {
      PDEVNODE pdn;
      pdevnode = pdevnode->next;
      pdn = pdevnode;          /* skip over header node */
      for (count=0 ; pdn != NULL ; pdn=pdn->next, count++);
    } else { /* ATOMIC */
      count = 1;
    }

    /* Allocate space for array and copy names */
    namearr = (devnamearr)malloc((count+1)*sizeof(char *));
    *names = namearr;

    for (i=0 ; i < count ; i++) {
      namearr[i] = (char *)malloc(255);
      strcpy(namearr[i],pdevnode->dname);
      pdevnode = pdevnode->next;
    }
  }
  namearr[count] = NULL;
  *cnt = count;
  return(DEV_NORMAL);
}


/*************************************************************************
 * FUNCTION : devInfo()
 * PURPOSE  : Retrieves structure(s) containing all information the
 *            device server has on the given device dname. An array of
 *            pointers to DEVNODE's is returned. If dname is an atomic
 *            device, there is one element in the array for each message
 *            the device supports. If dname is a composite device, there
 *            is an element for each message of each constituent atomic
 *            device. A NULL pointer indicates no more entries. Space 
 *            for the array is automatically allocated, and subsequently
 *            freed up on the next call. NOTE: pchid field of DEVNODE
 *            is not initialized by this call. Use devSend or devChid.
 * ARGS in  : dname: name of device (atomic or composite)
 * ARGS out : infoarr: array of pointers to DEVNODE's
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL - success, cnt is valid
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 *            DEV_NODEV    - dname does not exist
 ************************************************************************/
int devInfo(char *dname, devstructarr *infoarr)
{
  char *localname = dname; /* RPC call argument */
  resolve_result *result;  /* RPC call return   */
  devicelist reply,list;   /* lst part of RPC call return */
  int ndev, i;
  static devstructarr devarr = NULL;
  PDEVNODE pdevnode;       /* device node structure */

  if (first_call) {
    if (init_libdev() != DEV_NORMAL)
      return(DEV_NOSERVER);
    first_call = 0;
  }

  /* Contact device server to retrieve linked list of information */
  if ((result = device_info_1(&localname,cl)) == NULL) {
    clnt_perror(cl,server);
    return(DEV_CALLFAIL);
  }
  if (result->myerrno == NOSUCHDEV)
    return(DEV_NODEV);

  /* Extract ptr to device server response.
     Reply is actually a linked list. For atomic devices, there is
     one node for each message supported by that device. For composite
     devices, there are n+1 nodes for a composite device composed of
     n atomic devices. The first node indicates COMPOSITE in the type
     field.
  */
  reply = result->reply;
  if (reply == NULL) 
    return(DEV_CALLFAIL);

  /* Free up malloc'ed space from last call to devInfo */
  if (devarr != NULL) {
    int i = 0;
    while (devarr[i] != NULL) {
      free(devarr[i]);
      i++;
    }
    free(devarr);
  }

  /* Count up number of nodes by traversing list */
  list = reply;
  for (ndev = 0 ; list != NULL ; list=list->next, ndev++);
  devarr = (devstructarr)malloc((ndev+1)*sizeof(devstructarr *));
  *infoarr = devarr;

  /* Copy information into malloc'ed array */
  for (i=0 ; i < ndev ; i++) {
    devarr[i] = (PDEVNODE)malloc(sizeof(DEVNODE));
    copyReplynode(devarr[i],reply);
    reply = reply->next;
  }
  devarr[ndev] = NULL;   /* last element must be NULL */
  return(DEV_NORMAL);
}

/*************************************************************************
 * FUNCTION : devSend()
 * PURPOSE  : Sends msg to dname. Depending on definition of msg for the
 *            named device, the memory pointed to by arg may be read or
 *            written. If dname is not found in the local hash table, the
 *            device server will contacted to get the device definition.
 *            The device definition determines which PV to read/write
 *            for the given msg.
 * ARGS in  : dname: device name
 *            msg:   message
 *            arg:   will be read if msg results in a write to dname
 * ARGS out : arg:   will be written if msg results in a read from dname
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL - success, msg was sent to dname
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 *            DEV_NODEV    - dname does not exist
 *            DEV_NOMSG    - msg is not defined for dname
 *            DEV_BADSEARCH - unable to ca_search on PV
 *            DEV_BADCHID   - unable to put/get to PV
 *            DEV_TIMEOUT   - ca_pend_io operation timed out
 ************************************************************************/
int devSend(char *dname, char *msg, void *arg)
{
  PDEVNODE pdevnode;       /* ptr to userPvt of hash table entry */
  GPDENTRY *pentry;        /* hash table entry */
  int status;

  if (first_call) {
    if (init_libdev() != DEV_NORMAL)
      return(DEV_NOSERVER);
    first_call = 0;
  }

  /* Try to find dname in local hash table */
  pentry = gpdFind(table,dname,msg);

  if (pentry == NULL) {           /* need to resolve and add to hash table */
    status = addDevice(dname,msg);
    if (status != DEV_NORMAL && status != DEV_NOMSG)
      return(status);
    pentry = gpdFind(table,dname,msg);
  }
  pdevnode = (PDEVNODE)pentry->userPvt;

  /* Perform operation on device (what to do is specified in devnode) */
  status = msgOp(pdevnode,arg);
  return(status);
}

/*************************************************************************
 * FUNCTION : devSendBlock()
 * PURPOSE  : Sends msg to dname. Depending on definition of msg for the
 *            named device, the memory pointed to by arg may be read or
 *            written. If dname is not found in the local hash table, the
 *            device server will contacted to get the device definition.
 *            The device definition determines which PV to read/write
 *            for the given msg. The call will block until the record
 *            associated with the PV has completed processing. 
 * ARGS in  : dname: device name
 *            msg:   message
 *            arg:   will be read if msg results in a write to dname
 * ARGS out : arg:   will be written if msg results in a read from dname
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL - success, msg was sent to dname
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 *            DEV_NODEV    - dname does not exist
 *            DEV_NOMSG    - msg is not defined for dname
 *            DEV_BADSEARCH - unable to ca_search on PV
 *            DEV_BADCHID   - unable to put/get to PV
 *            DEV_TIMEOUT   - ca_pend_io operation timed out
 ************************************************************************/
int devSendBlock(char *dname, char *msg, void *arg)
{
  /* Must wait until blocking capability added to channel access */
  return(devSend(dname, msg, arg));
}

/*************************************************************************
 * FUNCTION : devChid()
 * PURPOSE  : Returns initialized chid for given atomic device and message.
 *            Provides a hook to Channel Access layer, most calls of which
 *            require a chid. In the event the chid becomes "disconnected",
 *            devChid() may be called again to re-establish the chid.
 * ARGS in  : dname: device name
 *            msg:   message
 * ARGS out : pchid: ptr to ptr to chid
 * GLOBAL   : May result in RPC client initialization.
 * RETURNS  : DEV_NORMAL    - success, msg was sent to dname
 *            DEV_ATOMICONLY - dname must be of type ATOMIC
 *            DEV_NOSERVER  - unable to contact device server
 *            DEV_CALLFAIL  - RPC call failed
 *            DEV_NODEV     - dname does not exist
 *            DEV_NOMSG     - msg not defined for dname
 *            DEV_BADSEARCH - unable to ca_search on PV
 *            DEV_BADCHID   - unable to put/get to PV
 *            DEV_TIMEOUT   - ca_pend_io operation timed out
 ************************************************************************/
int devChid(char *dname, char *msg, chid **pchid)
{
  PDEVNODE pdevnode;       /* ptr to userPvt of hash table entry */
  GPDENTRY *pentry;        /* hash table entry */
  int status;

  if (first_call) {
    if (init_libdev() != DEV_NORMAL)
      return(DEV_NOSERVER);
    first_call = 0;
  }

  /* Try to find dname in local hash table */
  pentry = gpdFind(table,dname,msg);

  if (pentry == NULL) {           /* need to resolve and add to hash table */
    status = addDevice(dname,msg);
    if (status != DEV_NORMAL)
      return(status);
    pentry = gpdFind(table,dname,msg);
  }
  pdevnode = (PDEVNODE)pentry->userPvt;  /* get user part of table entry */
  
  if (pdevnode->type == COMPOSITE) /* atomic only */
    return(DEV_ATOMICONLY);

  /* Check to make sure chid is connected. If not, try once. */
  if (ca_state((*(pdevnode->pchid))) != cs_conn) {
    status = ca_search(pdevnode->pvname,pdevnode->pchid);
    if (status != ECA_NORMAL)
      return(DEV_BADSEARCH);
    status = ca_pend_io(timeout);
    if (status != ECA_NORMAL)
      return(DEV_TIMEOUT);
  }
  
  *pchid = pdevnode->pchid;
  status = DEV_NORMAL;
  return(status);
}

/*************************************************************************
 * FUNCTION : devDirType()
 * PURPOSE  : Given the devstructarr returned by the devInfo() function,
 *            and a particular message, devDirType determines the direction
 *            of io (READ or WRITE) and the data type (DBR_DOUBLE or DBR_
 *            STRING) that would apply if the message were sent to the device.
 *            If unable to determine dir and type from info, a default of
 *            dir = READ and type = DBR_DOUBLE is given.
 *            
 * ARGS in  : msg: string containing message name
 *            info: structure returned by devInfo() function
 * ARGS out : dir:   READ or WRITE 
 *            type:  DBR_DOUBLE or DBR_STRING 
 * GLOBAL   : nothing
 * RETURNS  : DEV_NORMAL - success
 ************************************************************************/
int devDirType(char *msg, devstructarr info, int *dir, int *type)
{
  int i;

  *type = DBR_DOUBLE;   /* default */
  *dir  = READ;

  if (info == NULL)
    return(DEV_NORMAL);

  for (i=0 ; info[i] != NULL ; i++)     /* scan through info */
    if (!strcmp(info[i]->msg,msg)) {    /* until first matching msg found */
      *type = info[i]->datatype;
      *dir = info[i]->dir;
    }

  return(DEV_NORMAL);
}

/*************************************************************************
 * FUNCTION : addDevice()
 * PURPOSE  : Contacts device server and retrieves device definition
 *            for dname. The device server's reply is copied into the
 *            local DEVNODE struct. Space is malloc'ed for a chid and
 *            an initial ca_search() is performed to initialize the chid.
 *            The device name and DEVNODE struct(s) are then added to the
 *            local hash table. 
 * ARGS in  : dname: device name
 *            msg:   message
 * ARGS out : none
 * GLOBAL   : Contacts device server. Adds to hash table.
 * RETURNS  : DEV_NORMAL - success, dname was added to hash table
 *            DEV_NOSERVER - unable to contact device server
 *            DEV_CALLFAIL - RPC call failed
 *            DEV_NODEV    - dname does not exist
 *            DEV_NOMSG    - msg is not defined for dname
 *            DEV_BADSEARCH - unable to ca_search on PV
 *            DEV_TIMEOUT   - ca_pend_io operation timed out
 ************************************************************************/
static int addDevice(char *dname, char *msg)
{
  requesttype request;               /* RPC call argument */
  resolve_result *result;            /* RPC call return   */
  devicelist reply;                  /* list part of RPC call return */
  PDEVNODE pdevnode, prevpdevnode;   /* device node structure */
  GPDENTRY *pentry;                  /* hash table entry */
  int status;
  int nohandle = 0;

  request.dname = malloc(80);      /* fill out request struct */
  request.mesg = malloc(80);
  strcpy(request.dname,dname);
  strcpy(request.mesg,msg);

  /* Contact device server and get device definition */
  if ((result = resolve_device_1(&request,cl)) == NULL) {
    clnt_perror(cl,server);
    return(DEV_CALLFAIL);
  }

  /* Return immediately if requested device doesn't exist */
  if (result->myerrno == NOSUCHDEV)
    return(DEV_NODEV);

  /* If device doesn't handle msg, we still must add a dummy node to table */
  if (result->myerrno == NOSUCHMSG)
    nohandle = 1;

  /* Extract ptr to device server response.
     Reply is actually a linked list. One node for atomic devices, and
     n+1 nodes for a composite device composed of n atomic devices.
  */
  reply = result->reply;
  if (reply == NULL)
    return(DEV_CALLFAIL);

  /* Copy out relevant information and add DEVNODE to hash table */
  pdevnode = (PDEVNODE)malloc(sizeof(DEVNODE));
  copyReplynode(pdevnode,reply);
  pentry = gpdAdd(table,pdevnode->dname,pdevnode->msg);
  pentry->userPvt = (void *)pdevnode;

  status = ECA_NORMAL;
  switch (reply->type) {     /* malloc chid(s) and initialize with search */
  case DUMMY: {
    return(DEV_NOMSG);
  }
  case ATOMIC: {         
    pdevnode->pchid = (chid *)malloc(sizeof(chid));
    status = ca_search(pdevnode->pvname,pdevnode->pchid);
    status = ca_pend_io(timeout);
    if (status == ECA_NORMAL)
      return(DEV_NORMAL);
    else
      return(DEV_BADSEARCH);
  }
  case COMPOSITE: {          
    prevpdevnode = pdevnode;
    reply = reply->next; 
    while (reply != NULL) {
      pdevnode = (PDEVNODE)malloc(sizeof(DEVNODE));
      copyReplynode(pdevnode,reply);
      pentry = gpdFind(table,pdevnode->dname,pdevnode->msg);
      if (pentry == NULL && pdevnode->type == DUMMY) {
	pentry = gpdAdd(table,pdevnode->dname,pdevnode->msg);
	pentry->userPvt = (void *)pdevnode;
      }
      else if (pentry == NULL) {
	pdevnode->pchid = (chid *)malloc(sizeof(chid));
	pentry = gpdAdd(table,pdevnode->dname,pdevnode->msg);
	pentry->userPvt = (void *)pdevnode;
	status = ca_search(pdevnode->pvname,pdevnode->pchid);
	if (status != ECA_NORMAL)
	  break;
      }
      else {   /* if atomic device already in hash table, use its chid */
	pdevnode->pchid = ((PDEVNODE)pentry->userPvt)->pchid;
      }
      prevpdevnode->next = pdevnode;
      prevpdevnode = pdevnode;
      reply = reply->next;
    }
    if (status != ECA_NORMAL)
      return(DEV_BADSEARCH);
    status = ca_pend_io(timeout);
    if (status == ECA_NORMAL) {
      if (nohandle)
	return(DEV_NOMSG);
      else
	return(DEV_NORMAL);
    }
    else
      return(DEV_BADSEARCH);
  }
  default:
    return(DEV_NODEV);
  }
}

/*************************************************************************
 * FUNCTION : msgOp()
 * PURPOSE  : Reads or writes PV(s) specified in devnode. The memory
 *            pointed to by arg is the source/sink for the read/write.
 *            The devnode contains the direction (r/w), PV, and data
 *            type for the channel access operation.
 * ARGS in  : pdevnode: ptr to a linked list of device nodes
 *            arg: memory is read from if devnode specifies a read
 * ARGS out : arg: memory is written to if devnode specified a write
 * GLOBAL   : 
 * RETURNS  : DEV_NORMAL - success, PV written/read
 *            DEV_BADSEARCH - unable to establish connection with PV
 *            DEV_BADCHID   - unable to put/get to PV
 *            DEV_TIMEOUT   - ca_pend_io operation timed out
 ************************************************************************/
static int msgOp(PDEVNODE pdevnode,void *arg)
{
  PDEVNODE pnode = pdevnode;
  int status;
  void *parg = arg;
  static double defval;
  int nohandle = 0;

  /* Make sure channels are connected. Return if unable to connect. */
  switch (pnode->type) {
  case ATOMIC: {
    if (ca_state((*(pnode->pchid))) != cs_conn) {
      status = ca_search(pnode->pvname,pnode->pchid);
      if (status != ECA_NORMAL)
	return(DEV_BADSEARCH);
      status = ca_pend_io(timeout);
      if (status != ECA_NORMAL)
	return(DEV_TIMEOUT);
    }
    break;
  }
  case COMPOSITE: {
    pnode = pnode->next;   /* skip over header node */
    while (pnode != NULL) {
      if ((pnode->type != DUMMY) && (ca_state((*(pnode->pchid))) != cs_conn)) {
	status = ca_search(pnode->pvname,pnode->pchid);
	if (status != ECA_NORMAL)    /* return on first bad connection */
	  return(DEV_BADSEARCH);
	status = ca_pend_io(timeout);
	if (status != ECA_NORMAL)
	  return(DEV_TIMEOUT);
      }
      pnode = pnode->next;
    }
    break;
  }
  case DUMMY: { break; }  /* device doesn't handle this message */
  }

  pnode = pdevnode;
  /* Assume channels are connected at this point. Perform get/put. */
  switch (pnode->type) {

  case DUMMY: {   /* device doesn't handle this message */
    parg = configure_parg(pnode->datatype, pnode->dir, pnode->nargs,
			  pnode->def, pnode->type, arg);
    if (pnode->dir == READ)
      switch (pnode->datatype) {
      case DBR_DOUBLE: {
	*(double *)parg = quiet_nan(0);
	break;
      }
      case DBR_STRING: {
	*(char **)parg = NULL;
	break;
      }
      }
    return(DEV_NOMSG);
  }
  case ATOMIC: {
    parg = configure_parg(pnode->datatype, pnode->dir, pnode->nargs,
			  pnode->def, pnode->type, arg);
    switch (pnode->dir) {
    case READ: {
      status = ca_array_get(pnode->datatype,
			    ca_element_count(*(pnode->pchid)),
			    *(pnode->pchid),parg);
      break;
    }
    case WRITE: {
      status = ca_array_put(pnode->datatype,
			    ca_element_count(*(pnode->pchid)),
			    *(pnode->pchid),parg);
      break;
    }
    }
    if (status != ECA_NORMAL)
      return(DEV_BADCHID);
    status = ca_pend_io(timeout);
    if (status != ECA_NORMAL)
      return(DEV_TIMEOUT);
    break;
  }

  case COMPOSITE: {    
    pnode = pnode->next;  /* skip over header node */
    while (pnode != NULL) {
      switch (pnode->dir) {
      case READ: {
	parg = configure_parg(pnode->datatype, pnode->dir, pnode->nargs,
			      pnode->def, pnode->type, arg);
	switch (pnode->datatype) {
	case DBR_DOUBLE: {
	  if (pnode->type == DUMMY) {
	    nohandle = 1;
	    *(double *)parg = quiet_nan(0);
	    status = ECA_NORMAL; /* fake it */
	  }
	  else {
	    status = ca_array_get(pnode->datatype,
				  ca_element_count(*(pnode->pchid)),
				  *(pnode->pchid),parg);	    
	  }
	  break;
	}
	case DBR_STRING: {
	  if (pnode->type == DUMMY) {
	    nohandle = 1;
	    *(char **)parg = NULL;
	    status = ECA_NORMAL; /* fake it */
	  }
	  else {
	    status = ca_array_get(pnode->datatype,
				  ca_element_count(*(pnode->pchid)),
				  *(pnode->pchid),parg);
	  }
	  break;
	}
	}
	break;
      }
      case WRITE: {
	if (pnode->type != DUMMY) {
	  parg = configure_parg(pnode->datatype, pnode->dir, pnode->nargs,
				pnode->def, pnode->type, arg);
	  status = ca_array_put(pnode->datatype,
				ca_element_count(*(pnode->pchid)),
				*(pnode->pchid),parg);	  
	}
	break;
      }
      }
      
      if (status != ECA_NORMAL)
	return(DEV_BADCHID);
      status = ca_pend_io(timeout);
      if (status != ECA_NORMAL)
	return(DEV_TIMEOUT);
      if (pnode->nargs != 0)
	switch (pnode->datatype) {
	case DBR_DOUBLE: {
	  if (pnode->type == DUMMY)
	    arg = (double *)arg + 1;
	  else
	    arg = (double *)arg + ca_element_count(*(pnode->pchid));
	  break;
	}
	case DBR_STRING: {
	  arg = (char **)arg + 1;
	  break;
	}
	}
      pnode = pnode->next;
    } /* end while */
    break;
  }
  } 
  if (nohandle)
    return(DEV_NOMSG);
  else
    return(DEV_NORMAL);
}

/*************************************************************************
 * FUNCTION : devError()
 * PURPOSE  : Prints error message to stderr based on return code from
 *            the public functions of libdev.
 * ARGS in  : status: error code returned by public libdev function
 * ARGS out : none
 * GLOBAL   : Writes to stderr.
 * RETURNS  :  0 if status == DEV_NORMAL
 *            -1 if status != DEV_NORMAL
 ************************************************************************/
int devError(int status)
{
  if (status == DEV_NORMAL)
    return(0);
  else {
    fprintf(stderr,"device error: ");
    switch(status) {
    case DEV_NODEV: {
      fprintf(stderr,"no such device\n");
      break;
    }
    case DEV_NOMSG: {
      fprintf(stderr,"message undefined for device\n");
      break;
    }
    case DEV_CALLFAIL: {
      fprintf(stderr,"server call failed\n");
      break;
    }
    case DEV_NOSERVER: {
      fprintf(stderr,"unable to locate device server\n");
      break;
    }
    case DEV_BADCHID: {
      fprintf(stderr,"channel access put/get failed\n");
      break;
    }
    case DEV_TIMEOUT: {
      fprintf(stderr,"channel access operation timed out\n");
      break;
    }
    case DEV_BADSEARCH: {
      fprintf(stderr,"channel access search failed\n");
      break;
    }
    case DEV_ATOMICONLY: {
      fprintf(stderr,"this function restricted to atomic devices only\n");
      break;
    }
    default: {
      fprintf(stderr,"unknown error code\n");
      break;
    }
    }
    return(-1);
  }
}

/*************************************************************************
 * FUNCTION : devErrorString()
 * PURPOSE  : Copies an error string to buf which corresponds to given
 *            status code. 
 * ARGS in  : status: error code returned by public libdev function
 * ARGS out : buf: sufficient (80 chars) space in which to place message
 * GLOBAL   : 
 * RETURNS  :  0 if status == DEV_NORMAL
 *            -1 if status != DEV_NORMAL
 ************************************************************************/
int devErrorString(int status, char *buf)
{
  if (status == DEV_NORMAL)
    return(0);
  else {
    strcpy(buf,"device error: ");
    switch(status) {
    case DEV_NODEV: {
      strcat(buf,"no such device");
      break;
    }
    case DEV_NOMSG: {
      strcat(buf,"message undefined for device");
      break;
    }
    case DEV_CALLFAIL: {
      strcat(buf,"server call failed");
      break;
    }
    case DEV_NOSERVER: {
      strcat(buf,"unable to locate device server");
      break;
    }
    case DEV_BADCHID: {
      strcat(buf,"channel access put/get failed");
      break;
    }
    case DEV_TIMEOUT: {
      strcat(buf,"channel access operation timed out");
      break;
    }
    case DEV_BADSEARCH: {
      strcat(buf,"channel access search failed");
      break;
    }
    case DEV_ATOMICONLY: {
      strcat(buf,"this function restricted to atomic devices only");
      break;
    }
    default: {
      strcat(buf,"unknown error code");
      break;
    }
    }
    return(-1);
  }
}

/*************************************************************************
 * FUNCTION : devTimeout()
 * PURPOSE  : Sets global timeout period for pending input/output
 *            operations. Ie. timeout value for ca_pend_io() calls.
 * ARGS in  : tout: timeout period in seconds 
 * ARGS out : none
 * GLOBAL   : Modifies library timeout variable.
 * RETURNS  :  0 always
 ************************************************************************/
int devTimeout(float tout)
{
  timeout = tout;
  return(0);
}

/*************************************************************************
 * FUNCTION : copyReplynode()
 * PURPOSE  : Copies device definition information from the RPC reply
 *            to a local devnode struct.
 * ARGS in  : reply: pointer to a single device definition reply
 * ARGS out : pdevnode: copies info into struct 
 * GLOBAL   : 
 ************************************************************************/
static void copyReplynode(PDEVNODE pdevnode,devicelist reply)
{
  strcpy(pdevnode->dname,reply->dname);
  pdevnode->type = reply->type;
  strcpy(pdevnode->msg,reply->msg);
  strcpy(pdevnode->pvname,reply->pvname);
  pdevnode->dir = reply->dir;
  pdevnode->nargs = reply->nargs;
  pdevnode->datatype = reply->datatype;
  strcpy(pdevnode->def,reply->def);
  pdevnode->pchid = NULL;
  pdevnode->next = NULL;
}

/*************************************************************************
 * FUNCTION : connect_to_server()
 * PURPOSE  : Establishes RPC client connection to device server.
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : Initializes local hostname and gets server ident.
 * RETURNS  :  0 - success
 *            -1 - failed to find DEVICE_PROG service on local host
 ************************************************************************/
static int connect_to_server()
{
  gethostname(server,MAXHOSTNAMELEN);  /* get name of our local machine */
  if ((cl = clnt_create(server,DEVICE_PROG,DEVICE_VERS,"tcp")) == NULL) {
    clnt_pcreateerror(server);
    return(-1);
  }
  return(0);
}

/*************************************************************************
 * FUNCTION : init_libdev()
 * PURPOSE  : Initializes libdev. 
 * ARGS in  : none
 * ARGS out : none
 * GLOBAL   : Establishes RPC connection, initializes CA and hash table.
 * RETURNS  : DEV_NORMAL - libdev successfully initialized
 *            DEV_NOSERVER - unable to connect to device server
 ************************************************************************/
static int init_libdev()
{
    if (connect_to_server() == -1)
      return(DEV_NOSERVER);
    ca_task_initialize();    /* init CA */
    gpdInitPvt(&table);      /* init hash table */
    return(DEV_NORMAL);
}

/*************************************************************************
 * FUNCTION : configure_parg()
 * PURPOSE  : This function attempts to manage some of the difficulties 
 *            inherent in supporting multiple data types (DOUBLE and STRING).
 *            Given the void pointer (arg) supplied by the user, and 
 *            complete information regarding the device and message,
 *            configure another void pointer to be used in a ca_get/
 *            ca_put call. 
 * ARGS in  : datatype: DBR_DOUBLE or DBR_STRING
 *            dir:      READ or WRITE
 *            nargs:    0 (no arguments) 1 (user supplies arguments)
 *            def:      default value if nargs is 0
 *            type:     ATOMIC, COMPOSITE, or DUMMY
 *            arg:      pointer to user memory (to be read or written)
 * ARGS out : none
 * GLOBAL   : 
 * RETURNS  : Correctly configured pointer for use in ca_get/ca_put call.
 ************************************************************************/
static void *configure_parg(int datatype, int dir, int nargs,
			    char *def, int type, void *arg) {
  static double defval;
  static char *recordstr;
  void *lparg;
/*  double atof();*/

  switch (dir) {
  case READ: {
    lparg = arg;
    switch (datatype) {
    case DBR_DOUBLE: {     /* arg should be pointing to space for double */
      return(lparg);
    }
    case DBR_STRING: {
      if (type == DUMMY)   /* return ptr as is so higher level function can */
	return(lparg);     /*   set ptr to NULL, indicating invalid string  */
      else {               /* otherwise, malloc string and return ptr to it */
	recordstr = malloc(sizeof(char)*MAX_STRING_SIZE);
	*(char **)lparg = recordstr;
	return((void *)(*(char **)lparg));
      }
    }
    }
    break;
  }
  case WRITE: {
    switch (datatype) {
    case DBR_DOUBLE: {
      if (nargs == 0) {          /* use default, return pointer to it */
	defval = atof(def);
	return((void *)&defval);
      }
      else {
	return(arg);          /* arg should be pointing to space for double */
      }
    }
    case DBR_STRING: {
      if (nargs == 0)
	return((void *)def);
      else
	return((void *)(*(char **)arg));  /* dereference once */
    }
    }
    break;
  }
  }
}


/* Used by qsort() call in devMatch() function. Needed to dereference once
   for strcmp function */

static int stringcompare(const void *a, const void *b)
{
  char **astr, **bstr;
  astr = (char **)a;
  bstr = (char **)b;
  return(strcmp(*astr,*bstr));
}
